# Copyright (c) 2020 Mike Kinney

"""Mech integration tests: init from box file"""
import re
import subprocess


import pytest


@pytest.mark.int
def test_int_init_from_boxfile(helpers):
    """Test mech init from .box file."""

    test_dir = "tests/int/init_from_boxfile"
    helpers.cleanup_dir_and_vms_from_dir(test_dir)

    ubuntu = "ubuntu-18.04"
    box_file = "/tmp/{}.box".format(ubuntu)

    # download the file if we don't have it already
    # that way we "cache" the file
    commands = """
    if ! [ -f "{box_file}" ]; then
      wget -O "{box_file}" "https://vagrantcloud.com/bento/\
boxes/{ubuntu}/versions/201912.04.0/providers/vmware_desktop.box"
    fi
    """.format(box_file=box_file, ubuntu=ubuntu)
    results = subprocess.run(commands, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert results.returncode == 0

    # init from boxfile
    command = "mech init --box bento/{} file:{}".format(ubuntu, box_file)
    expected_lines = ["Initializing", "has been init"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # should start
    command = "mech up"
    expected_lines = ["Extracting", "Added network",
                      "Bringing machine", "Getting IP", "Sharing folders",
                      "started", "Provisioning"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # should be able to destroy
    command = "mech destroy -f"
    expected = "Deleting"
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    assert re.search(expected, stdout)



# Copyright (c) 2020 Mike Kinney

"""Mech integration tests: init from json file"""
import re
import subprocess


import pytest


@pytest.mark.int
def test_int_init_from_jsonfile(helpers):
    """Test mech init from .json file."""

    test_dir = "tests/int/init_from_jsonfile"
    helpers.cleanup_dir_and_vms_from_dir(test_dir)

    ubuntu = "ubuntu-18.04"

    jsonfile_contents = """{
    "description": "Bento Ubuntu box",
    "short_description": "ubuntu",
    "name": "bento/ubuntu-18.04",
    "versions": [
        {
            "version": "201912.04.0",
            "status": "active",
            "description_html": "Some html description",
            "description_markdown": "Some markdown description",
            "providers": [
                {
                    "name": "vmware_desktop",
                    "url":  "https://vagrantcloud.com/bento/boxes/ubuntu-18.04/\
versions/201912.04.0/providers/vmware_desktop.box",
                    "checksum": null,
                    "checksum_type": null
                }
            ]
        }
    ]
}"""

    jsonfile_name = "bento_1804.json"
    jsonfile_path = "{}/{}".format(test_dir, jsonfile_name)
    jsonfile_file = open(jsonfile_path, "w")
    jsonfile_file.write(jsonfile_contents)
    jsonfile_file.close()

    # init from jsonfile
    command = "mech init --box bento/{} file:{}".format(ubuntu, jsonfile_name)
    expected_lines = ["Initializing", "has been init"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # should start
    command = "mech up"
    expected_lines = ["could not be found", "vmware_desktop",
                      "integrity", "Extracting", "Added network",
                      "Bringing machine", "Getting IP", "Sharing folders",
                      "started", "Provisioning"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # should be able to destroy
    command = "mech destroy -f"
    expected = "Deleting"
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    assert re.search(expected, stdout)

    # clean up file
    helpers.cleanup_dir_and_vms_from_dir(test_dir)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016-2017 Kevin Chung
# Copyright (c) 2018 German Mendez Bravo (Kronuz)
# Copyright (c) 2020 Mike Kinney
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#

"""Mech utility functions."""

from __future__ import division, absolute_import

import os
import re
import random
import string
import sys
import json
import tarfile
import fnmatch
import logging
import tempfile
import textwrap
import subprocess
import collections
from shutil import copyfile, rmtree

import requests
from clint.textui import colored
from clint.textui import progress

from .vmrun import VMrun
import mech.vbm
from .compat import b2s, PY3, raw_input

LOGGER = logging.getLogger(__name__)


def main_dir():
    """Return the main directory."""
    return os.getcwd()


def mech_dir():
    """Return the mech directory."""
    return os.path.join(main_dir(), '.mech')


def makedirs(name, mode=0o777):
    """Make directories with mode supplied."""
    try:
        os.makedirs(name, mode)
    except OSError:
        pass


def confirm(prompt, default='y'):
    """Confirmation prompt."""
    default = default.lower()
    if default not in ['y', 'n']:
        default = 'y'
    choicebox = '[Y/n]' if default == 'y' else '[y/N]'
    prompt = prompt + ' ' + choicebox + ' '

    while True:
        some_input = raw_input(prompt).strip()
        if some_input == '':
            if default == 'y':
                return True
            else:
                return False

        if re.match('y(?:es)?', some_input, re.IGNORECASE):
            return True

        elif re.match('n(?:o)?', some_input, re.IGNORECASE):
            return False


def save_mechfile_entry(mechfile_entry, name, mechfile_should_exist=False):
    """Save the entry to the Mechfile."""
    LOGGER.debug('mechfile_entry:%s name:%s mechfile_should_exist:%s',
                 mechfile_entry, name, mechfile_should_exist)
    mechfile = load_mechfile(mechfile_should_exist)

    mechfile[name] = mechfile_entry

    LOGGER.debug("after adding name:%s mechfile:%s", name, mechfile)
    return save_mechfile(mechfile)


def remove_mechfile_entry(name, mechfile_should_exist=True):
    """Removed the entry from the Mechfile."""
    LOGGER.debug('name:%s mechfile_should_exist:%s', name, mechfile_should_exist)
    mechfile = load_mechfile(mechfile_should_exist)

    if mechfile.get(name):
        del mechfile[name]

    LOGGER.debug("after removing name:%s mechfile:%s", name, mechfile)
    return save_mechfile(mechfile)


def save_mechfile(mechfile):
    """Save the mechfile object (which is a dict) to a file called 'Mechfile'.
       Return True if save was successful.
    """
    LOGGER.debug('mechfile:%s', mechfile)
    with open(os.path.join(main_dir(), 'Mechfile'), 'w+') as the_file:
        json.dump(mechfile, the_file, sort_keys=True, indent=2, separators=(',', ': '))
    return True


def locate(path, glob):
    """Locate a file in the path provided."""
    for root, _, filenames in os.walk(path):
        for filename in filenames:
            if fnmatch.fnmatch(filename, glob):
                return os.path.abspath(os.path.join(root, filename))


def parse_vmx(path):
    """Parse the virtual machine configuration (.vmx) file and return an
       ordered dictionary.
    """
    vmx = collections.OrderedDict()
    with open(path) as the_file:
        for line in the_file:
            line = line.strip().split('=', 1)
            if len(line) > 1:
                vmx[line[0].rstrip()] = line[1].lstrip()
    return vmx


def update_vmx(path, numvcpus=None, memsize=None, no_nat=False):
    """Update the virtual machine configuration (.vmx)
       file with desired settings.
    """
    updated = False

    vmx = parse_vmx(path)

    # Check if there is an existing interface
    has_network = False
    for vmx_key in vmx:
        if vmx_key.startswith('ethernet'):
            has_network = True

    # Write one if there is not
    if not has_network:
        vmx["ethernet0.addresstype"] = "generated"
        vmx["ethernet0.bsdname"] = "en0"
        if not no_nat:
            vmx["ethernet0.connectiontype"] = "nat"
        vmx["ethernet0.displayname"] = "Ethernet"
        vmx["ethernet0.linkstatepropagation.enable"] = "FALSE"
        vmx["ethernet0.pcislotnumber"] = "32"
        vmx["ethernet0.present"] = "TRUE"
        vmx["ethernet0.virtualdev"] = "e1000"
        vmx["ethernet0.wakeonpcktrcv"] = "FALSE"
        print(colored.yellow("Added network interface to vmx file"))
        updated = True

    # write out vmx file if memsize or numvcpus was specified
    if numvcpus is not None:
        vmx["numvcpus"] = '"{}"'.format(numvcpus)
        updated = True

    if memsize is not None:
        vmx["memsize"] = '"{}"'.format(memsize)
        updated = True

    if updated:
        with open(path, 'w') as new_vmx:
            for key in vmx:
                value = vmx[key]
                row = "{} = {}".format(key, value)
                new_vmx.write(row + os.linesep)


def load_mechfile(should_exist=True):
    """Load the Mechfile from disk and return the mechfile as a dictionary."""
    mechfile_fullpath = os.path.join(main_dir(), 'Mechfile')
    LOGGER.debug("mechfile_fullpath:%s", mechfile_fullpath)
    if os.path.isfile(mechfile_fullpath):
        with open(mechfile_fullpath) as the_file:
            try:
                mechfile = json.loads(the_file.read())
                LOGGER.debug('mechfile:%s', mechfile)
                return mechfile
            except ValueError:
                print(colored.red("Invalid Mechfile." + os.linesep))
                return {}
    else:
        if should_exist:
            sys.exit(colored.red(textwrap.fill(
                     "Could not find a Mechfile in the current directory. "
                     "A Mech environment is required to run this command. Run `mech init` "
                     "to create a new Mech environment. Or specify the name of the VM you would "
                     "like to start with `mech up <name>`. A final option is to change to a "
                     "directory with a Mechfile and to try again.")))
        else:
            return {}


def default_shared_folders():
    """Return the default shared folders config.
       The host_path value of "../.." is because it is relative to the vmx file.
    """
    return [{'share_name': 'mech', 'host_path': '../..'}]


def build_mechfile_entry(location, box=None, name=None, box_version=None,
                         shared_folders=None, provider=None):
    """Build the Mechfile from the inputs."""
    LOGGER.debug("location:%s name:%s box:%s box_version:%s provider:%s",
                 location, name, box, box_version, provider)
    mechfile_entry = {}

    if location is None:
        return mechfile_entry

    mechfile_entry['name'] = name
    mechfile_entry['box'] = box
    mechfile_entry['box_version'] = box_version
    mechfile_entry['provider'] = provider

    if shared_folders is None:
        shared_folders = default_shared_folders()
    mechfile_entry['shared_folders'] = shared_folders

    if any(location.startswith(s) for s in ('https://', 'http://', 'ftp://')):
        if not name:
            name = 'first'
        mechfile_entry['url'] = location
        return mechfile_entry

    elif location.startswith('file:') or os.path.isfile(re.sub(r'^file:(?://)?', '', location)):
        if not name:
            name = 'first'
        location = re.sub(r'^file:(?://)?', '', location)
        LOGGER.debug('location:%s', location)
        mechfile_entry['file'] = location
        try:
            # see if the location/file is a json file
            with open(location) as the_file:
                # if an exception is not thrown, then set values and continue
                # to the end of the function
                catalog = json.loads(the_file.read())
                LOGGER.debug('catalog:%s', catalog)
        except (json.decoder.JSONDecodeError, ValueError):
            # this means the location/file is probably a .box file
            # or the json is invalid
            return mechfile_entry
        except IOError:
            # cannot open file
            sys.exit('Error: Cannot open file:({})'.format(location))
    else:
        try:
            account, box, ver = (location.split('/', 2) + ['', ''])[:3]
            if not account or not box:
                sys.exit(colored.red("Provided box name is not valid"))
            if ver:
                box_version = ver
            print(
                colored.blue("Loading metadata for box '{}'{}".format(
                    location, " ({})".format(box_version) if box_version else "")))
            url = 'https://app.vagrantup.com/{}/boxes/{}'.format(account, box)
            response = requests.get(url)
            response.raise_for_status()
            catalog = response.json()
        except (requests.HTTPError, ValueError) as exc:
            sys.exit(colored.red("Bad response from HashiCorp's Vagrant Cloud API: %s" % exc))
        except requests.ConnectionError:
            sys.exit(colored.red("Couldn't connect to HashiCorp's Vagrant Cloud API"))

    LOGGER.debug("catalog:%s name:%s box_version:%s", catalog, name, box_version)
    return catalog_to_mechfile(catalog, name=name, box=box,
                               box_version=box_version, provider=provider)


def catalog_to_mechfile(catalog, name=None, box=None, box_version=None, provider=None):
    """Convert the Hashicorp cloud catalog entry to Mechfile entry."""
    LOGGER.debug('catalog:%s name:%s box:%s box_version:%s', catalog, name, box, box_version)
    if provider is None:
        provider = 'vmware'
    mechfile = {}
    versions = catalog.get('versions', [])
    for ver in versions:
        current_version = ver['version']
        if not box_version or current_version == box_version:
            for a_provider in ver['providers']:
                if provider in a_provider['name']:
                    mechfile['name'] = name
                    mechfile['provider'] = provider
                    mechfile['box'] = catalog['name']
                    mechfile['box_version'] = current_version
                    mechfile['url'] = a_provider['url']
                    mechfile['shared_folders'] = default_shared_folders()
                    return mechfile
    sys.exit(colored.red("Couldn't find a compatible VM using catalog:{}".format(catalog)))


def tar_cmd(*args, **kwargs):
    """Build the tar command to be used to extract the box."""
    try:
        startupinfo = None
        if os.name == "nt":
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.SW_HIDE | subprocess.STARTF_USESHOWWINDOW
        # run the 'tar --help' command to see what capabilities are available
        proc = subprocess.Popen(['tar', '--help'], stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE, startupinfo=startupinfo)
    except OSError:
        return None
    if proc.returncode:
        return None
    stdoutdata, _ = map(b2s, proc.communicate())
    tar = ['tar']
    # if the tar_cmd has the option enabled *and* the capability was in the 'tar --help'
    # then append that option to the tar command (which is a list of strings)
    if kwargs.get('wildcards') and re.search(r'--wildcards\b', stdoutdata):
        tar.append('--wildcards')
    if kwargs.get('force_local') and re.search(r'--force-local\b', stdoutdata):
        tar.append('--force-local')
    if kwargs.get('fast_read') and sys.platform.startswith('darwin'):
        tar.append('--fast-read')
    tar.extend(args)
    return tar


def init_box(name, box=None, box_version=None, location=None, force=False, save=True,
             instance_path=None, numvcpus=None, memsize=None, no_nat=False, provider=None):
    """Initialize the box. This includes uncompressing the files
       from the box file and updating the vmx file with
       desired settings (if vmware).

       Return the full path to the vmx or vbox file.

       VMware will just use the files as extracted.
       VirtualBox needs to "import" the ovf. It creates a .vbox file.
    """
    LOGGER.debug("name:%s box:%s box_version:%s location:%s provider:%s",
                 name, box, box_version, location, provider)
    if provider is None:
        provider = 'vmware'

    look_for = '*.vmx'
    vbox_path = None
    instance_path_save = instance_path
    if provider == 'virtualbox':
        look_for = '*.ovf'
        vbox_path = locate(instance_path, '*.vbox')
        # need to extract the files somewhere, then we will
        # remove them after importing to virtualbox
        if vbox_path != '':
            instance_path += '_tmp'

    # if we do not find the vmx file nor is the already imported files in place
    found_vmx_or_ovf = locate(instance_path, look_for)
    if not found_vmx_or_ovf and vbox_path != '':
        name_version_box = add_box(
            name=name,
            box=box,
            box_version=box_version,
            location=location,
            force=force,
            save=save,
            provider=provider)
        if not name_version_box:
            sys.exit(colored.red("Cannot find a valid box with a VMX/OVF file in boxfile"))

        box_parts = box.split('/')
        box_dir = os.path.join(*filter(None, (mech_dir(), 'boxes', provider,
                                              box_parts[0], box_parts[1], box_version)))
        box_file = locate(box_dir, '*.box')

        print(colored.blue("Extracting box '{}'...".format(box_file)))
        makedirs(instance_path)
        if sys.platform == 'win32':
            cmd = tar_cmd('-xf', box_file, force_local=True)
        else:
            cmd = tar_cmd('-xf', box_file)
        if cmd:
            startupinfo = None
            if os.name == "nt":
                startupinfo = subprocess.STARTUPINFO()
                startupinfo.dwFlags |= subprocess.SW_HIDE | subprocess.STARTF_USESHOWWINDOW
            proc = subprocess.Popen(cmd, cwd=instance_path, startupinfo=startupinfo)
            if proc.wait():
                sys.exit(colored.red("Cannot extract box"))
        else:
            tar = tarfile.open(box_file, 'r')
            tar.extractall(instance_path)

        if not save and box.startswith(tempfile.gettempdir()):
            os.unlink(box)

    if provider == 'vmware':
        vmx_path = locate(instance_path, '*.vmx')
        if not vmx_path:
            sys.exit(colored.red("Cannot locate a VMX file"))
        update_vmx(vmx_path, numvcpus=numvcpus, memsize=memsize, no_nat=no_nat)
        return vmx_path
    else:
        ovf_path = locate(instance_path, '*.ovf')
        if not ovf_path:
            sys.exit(colored.red("Cannot locate an OVF file"))
        LOGGER.debug('ovf_path:%s', ovf_path)
        vbm = mech.vbm.VBoxManage()
        import_results = vbm.importvm(path_to_ovf=ovf_path, name=name,
                                      base_folder=mech_dir(), quiet=True)
        LOGGER.debug('import_results:%s', import_results)
        vbox_path = locate(instance_path_save, '*.vbox')
        if not vbox_path:
            sys.exit(colored.red("Cannot locate an vbox file"))
        # remove the extracted files
        rmtree(instance_path)
        return vbox_path


def add_box(name=None, box=None, box_version=None, location=None,
            force=False, save=True, provider=None):
    """Add a box."""
    # build the dict
    LOGGER.debug('name:%s box:%s box_version:%s location:%s provider:%s',
                 name, box, box_version, location, provider)

    mechfile_entry = build_mechfile_entry(
        box=box,
        name=name,
        location=location,
        box_version=box_version,
        provider=provider)

    return add_mechfile(
        mechfile_entry,
        name=name,
        box=box,
        location=location,
        box_version=box_version,
        force=force,
        save=save,
        provider=provider)


def add_mechfile(mechfile_entry, name=None, box=None, box_version=None,
                 location=None, force=False, save=True, provider=None):
    """Add a mechfile entry."""
    LOGGER.debug('mechfile_entry:%s name:%s box:%s box_version:%s location:%s provider:%s',
                 mechfile_entry, name, box, box_version, location, provider)

    box = mechfile_entry.get('box')
    name = mechfile_entry.get('name')
    box_version = mechfile_entry.get('box_version')
    provider = mechfile_entry.get('provider')

    url = mechfile_entry.get('url')
    box_file = mechfile_entry.get('file')

    if box_file:
        return add_box_file(box=box, box_version=box_version, filename=box_file,
                            force=force, save=save, provider=provider)

    if url:
        return add_box_url(name=name, box=box, box_version=box_version,
                           url=url, force=force, save=save, provider=provider)
    print(
        colored.red(
            "Could not find a VMWare compatible VM for '{}'{}".format(
                name, " ({})".format(box_version) if box_version else "")))


def add_box_url(name, box, box_version, url, force=False, save=True, provider=None):
    """Add a box using the URL."""
    LOGGER.debug('name:%s box:%s box_version:%s url:%s provider:%s',
                 name, box, box_version, url, provider)
    boxname = os.path.basename(url)
    box_parts = box.split('/')
    first_box_part = box_parts[0]
    second_box_part = ''
    if len(box_parts) > 1:
        second_box_part = box_parts[1]
    if provider is None:
        provider = 'vmware'
    box_dir = os.path.join(*filter(None, (mech_dir(), 'boxes', provider,
                                          first_box_part, second_box_part, box_version)))
    exists = os.path.exists(box_dir)
    if not exists or force:
        if exists:
            print(colored.blue("Attempting to download provider:{} box:'{}'...".format(provider,
                                                                                       box)))
        else:
            print(colored.blue("Provider:{} Box:'{}' could not be found. "
                               "Attempting to download...".format(provider, box)))
        try:
            print(colored.blue("URL: {}".format(url)))
            response = requests.get(url, stream=True)
            response.raise_for_status()
            try:
                length = int(response.headers['content-length'])
                progress_args = dict(expected_size=length // 1024 + 1)
                progress_type = progress.bar
            except KeyError:
                progress_args = dict(every=1024 * 100)
                progress_type = progress.dots
            the_file = tempfile.NamedTemporaryFile(delete=False)
            try:
                for chunk in progress_type(
                        response.iter_content(
                            chunk_size=1024),
                        label="{} ".format(boxname),
                        **progress_args):
                    if chunk:
                        the_file.write(chunk)
                the_file.close()
                if response.headers.get('content-type') == 'application/json':
                    # Downloaded URL might be a Vagrant catalog if it's json:
                    catalog = json.load(the_file.name)
                    mechfile = catalog_to_mechfile(catalog, name, box, box_version)
                    return add_mechfile(
                        mechfile,
                        name=name,
                        box_version=box_version,
                        force=force,
                        save=save)
                else:
                    # Otherwise it must be a valid box:
                    return add_box_file(box=box, box_version=box_version,
                                        filename=the_file.name, url=url, force=force,
                                        save=save, provider=provider)
            finally:
                os.unlink(the_file.name)
        except requests.HTTPError as exc:
            sys.exit(colored.red("Bad response: %s" % exc))
        except requests.ConnectionError:
            sys.exit(colored.red("Couldn't connect to '%s'" % url))
    return name, box_version, box


def add_box_file(box=None, box_version=None, filename=None, url=None,
                 force=False, save=True, provider=None):
    """Add a box using a file as the source. Returns box and box_version."""
    print(colored.blue("Checking integrity of provider:{} box:'{}' "
                       "filename:{}...".format(provider, box, filename)))

    valid_endswith = 'vmx'
    look_for = '*.vmx'
    if provider == 'virtualbox':
        valid_endswith = 'ovf'
        look_for = '*.ovf'

    if sys.platform == 'win32':
        cmd = tar_cmd('-tf', filename, look_for, wildcards=True, fast_read=True, force_local=True)
    else:
        cmd = tar_cmd('-tf', filename, look_for, wildcards=True, fast_read=True)

    if cmd:
        startupinfo = None
        if os.name == "nt":
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.SW_HIDE | subprocess.STARTF_USESHOWWINDOW
        proc = subprocess.Popen(cmd, startupinfo=startupinfo)
        valid_tar = not proc.wait()
    else:
        tar = tarfile.open(filename, 'r')
        files = tar.getnames()
        valid_tar = False
        for i in files:
            if i.endswith(valid_endswith):
                valid_tar = True
                break
            if i.startswith('/') or i.startswith('..'):
                sys.exit(colored.red(textwrap.fill(
                         "This box is comprised of filenames starting with '/' or '..' "
                         "Exiting for the safety of your files.")))

    if valid_tar:
        if save:
            boxname = os.path.basename(url if url else filename)
            box = os.path.join(*filter(None, (mech_dir(), 'boxes', provider, box,
                                              box_version, boxname)))
            path = os.path.dirname(box)
            makedirs(path)
            if not os.path.exists(box) or force:
                copyfile(filename, box)
        else:
            box = filename
        return box, box_version


def get_info_for_auth(mech_use=False):
    """Get information (username/pub_key) for authentication."""
    username = os.getlogin()
    pub_key = os.path.expanduser('~/.ssh/id_rsa.pub')
    return {'auth': {'username': username, 'pub_key': pub_key, 'mech_use': mech_use}}


def init_mechfile(location=None, box=None, name=None, box_version=None, add_me=None,
                  use_me=None, provider=None):
    """Initialize the Mechfile."""
    LOGGER.debug("name:%s box:%s box_version:%s location:%s add_me:%s use_me:%s provider:%s",
                 name, box, box_version, location, add_me, use_me, provider)
    mechfile_entry = build_mechfile_entry(
        location=location,
        box=box,
        name=name,
        box_version=box_version,
        provider=provider)
    if add_me:
        mechfile_entry.update(get_info_for_auth(use_me))
    LOGGER.debug('mechfile_entry:%s', mechfile_entry)
    return save_mechfile_entry(mechfile_entry, name, mechfile_should_exist=False)


def add_to_mechfile(location=None, box=None, name=None, box_version=None, add_me=None,
                    use_me=None, provider=None):
    """Add entry to the Mechfile."""
    LOGGER.debug("name:%s box:%s box_version:%s location:%s add_me:%s use_me:%s provider:%s",
                 name, box, box_version, location, add_me, use_me, provider)
    this_mech_entry = build_mechfile_entry(
        location=location,
        box=box,
        name=name,
        box_version=box_version,
        provider=provider)
    if add_me:
        this_mech_entry.update(get_info_for_auth(use_me))
    LOGGER.debug('this_mech_entry:%s', this_mech_entry)
    return save_mechfile_entry(this_mech_entry, name, mechfile_should_exist=False)


def random_string(string_len=15):
    """Generate a random string of fixed length."""
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(string_len))


def add_auth(instance):
    """Add authentication to VM."""

    if not instance:
        sys.exit(colored.red("Need to provide an instance to add_auth()."))

    if instance.vmx is None:
        sys.exit(colored.red("Need to provide vmx add_auth()."))

    if instance.user is None or instance.user == '':
        sys.exit(colored.red("Need to provide user to add_auth()."))

    if instance.password is None or instance.password == '':
        sys.exit(colored.red("Need to provide password to add_auth()."))

    print(colored.green('Adding auth to instance:{}'.format(instance.name)))

    vmrun = VMrun(instance.vmx, instance.user, instance.password)
    # cannot run if vmware tools are not installed
    if not vmrun.installed_tools():
        sys.exit(colored.red("Cannot add auth if VMware Tools are not installed."))

    if instance.auth:
        username = instance.auth.get('username', None)
        pub_key = instance.auth.get('pub_key', None)
        if username and pub_key:
            pub_key_path = os.path.expanduser(pub_key)
            LOGGER.debug("pub_key_path:%s pub_key:%s", pub_key_path, pub_key)
            with open(pub_key_path, 'r') as the_file:
                pub_key_contents = the_file.read().strip()
            if pub_key_contents:
                # set the password to some random string
                # user should never need it (sudo should not prompt for a
                # password)
                password = random_string()
                cmd = ('sudo useradd -m -s /bin/bash -p "{password}" {username};'
                       'sudo mkdir /home/{username}/.ssh;'
                       'sudo usermod -aG sudo {username};'
                       'echo "{username} ALL=(ALL) NOPASSWD: ALL" | '
                       'sudo tee -a /etc/sudoers;'
                       'echo "{pub_key_contents}" | '
                       'sudo tee -a /home/{username}/.ssh/authorized_keys;'
                       'sudo chmod 700 /home/{username}/.ssh;'
                       'sudo chown {username}:{username} /home/{username}/.ssh;'
                       'sudo chmod 600 /home/{username}/.ssh/authorized_keys;'
                       'sudo chown {username}:{username} /home/{username}/.ssh/authorized_keys'
                       ).format(username=username, pub_key_contents=pub_key_contents,
                                password=password)
                LOGGER.debug('cmd:%s', cmd)
                results = vmrun.run_script_in_guest('/bin/sh', cmd, quiet=True)
                LOGGER.debug('results:%s', results)
                if results is None:
                    print(colored.red("Did not add auth"))
                else:
                    print(colored.green("Added auth."))
            else:
                print(colored.green("Could not read contents of the pub_key"
                                    " file:{}".format(pub_key)))
        else:
            print(colored.blue("Warning: Need a username and pub_key in auth."))
    else:
        print(colored.blue("No auth to add."))


def ssh(instance, command, plain=None, extra=None):
    """Run ssh command.
       Note: May not really need the tempfile if self.use_psk==True.
             Using the tempfile, there are options to not add host to the known_hosts files
             which is useful, but could be MITM attacks. Not likely locally, but still
             could be an issue.
    """
    LOGGER.debug('command:%s plain:%s extra:%s', command, plain, extra)
    if instance.created:
        config_ssh = instance.config_ssh()
        temp_file = tempfile.NamedTemporaryFile(delete=False)
        try:
            temp_file.write(config_ssh_string(config_ssh).encode('utf-8'))
            temp_file.close()

            cmds = ['ssh']
            if not plain:
                cmds.extend(('-F', temp_file.name))
            if extra:
                cmds.extend(extra)
            if not plain:
                cmds.append(config_ssh['Host'])
            if command:
                cmds.extend(('--', command))

            LOGGER.debug(
                " ".join(
                    "'{}'".format(
                        c.replace(
                            "'",
                            "\\'")) if ' ' in c else c for c in cmds))

            # if running a script
            if command:
                result = subprocess.run(cmds, capture_output=True)
                stdout = result.stdout.decode('utf-8').strip()
                stderr = result.stderr.decode('utf-8').strip()
                return result.returncode, stdout, stderr
            else:
                # interactive
                return subprocess.call(cmds), None, None
        finally:
            os.unlink(temp_file.name)


def scp(instance, src, dst, dst_is_host, extra=None):
    """Run scp command.
       Note: May not really need the tempfile if self.use_psk==True.
             Using the tempfile, there are options to not add host to the known_hosts files
             which is useful, but could be MITM attacks. Not likely locally, but still
             could be an issue.
    """
    if instance.created:

        config_ssh = instance.config_ssh()
        temp_file = tempfile.NamedTemporaryFile(delete=False)

        try:
            temp_file.write(config_ssh_string(config_ssh).encode())
            temp_file.close()

            cmds = ['scp']
            cmds.extend(('-F', temp_file.name))
            if extra:
                cmds.extend(extra)

            host = config_ssh['Host']
            dst = '{}:{}'.format(host, dst) if dst_is_host else dst
            src = '{}:{}'.format(host, src) if not dst_is_host else src
            cmds.extend((src, dst))

            LOGGER.debug(
                " ".join(
                    "'{}'".format(
                        c.replace(
                            "'",
                            "\\'")) if ' ' in c else c for c in cmds))
            return subprocess.run(cmds, capture_output=True)
        finally:
            os.unlink(temp_file.name)


def del_user(instance, username):
    """Delete a user in guest VM."""

    if not instance:
        sys.exit(colored.red("Need to provide an instance to del_user()."))

    if instance.vmx is None:
        sys.exit(colored.red("VM must be created."))

    if instance.user is None:
        sys.exit(colored.red("A user is required."))

    if username is None or username == '':
        sys.exit(colored.red("A username to delete is required."))

    print(colored.green('Removing username ({}) from instance:{}...'.format(username,
                                                                            instance.name)))

    cmd = 'sudo userdel -fr vagrant'
    LOGGER.debug('cmd:%s', cmd)

    if instance.use_psk:
        ssh(instance, cmd)
    else:
        vmrun = VMrun(instance.vmx, user=instance.user,
                      password=instance.password)
        # cannot run if vmware tools are not installed
        if not vmrun.installed_tools():
            sys.exit(colored.red("Cannot delete user if VMware Tools are not installed."))
        results = vmrun.run_script_in_guest('/bin/sh', cmd, quiet=True)
        LOGGER.debug('results:%s', results)
        if results is None:
            print(colored.red("Failed deleting user."))
        else:
            print(colored.green("Successfully deleted user ({}).".format(username)))


def provision(instance, show=False):
    """Provision an instance.

    Args:
        instance (MechInstance): an instance
        show (bool): just print the provisioning

    Notes:
        Valid provision types are:
           file: copies files to instances
           shell: executes scripts

    """

    if not instance:
        sys.exit(colored.red("Need to provide an instance to provision()."))

    if instance.vmx is None or instance.user is None:
        sys.exit(colored.red("Need to provide vmx/user to provision()."))

    print(colored.green('Provisioning instance:{}'.format(instance.name)))

    vmrun = VMrun(instance.vmx, instance.user, instance.password)
    # cannot run provisioning if vmware tools are not installed
    if not vmrun.installed_tools():
        sys.exit(colored.red("Cannot provision if VMware Tools are not installed."))

    provisioned = 0
    if instance.provision:
        for i, pro in enumerate(instance.provision):
            provision_type = pro.get('type')
            if provision_type == 'file':
                source = pro.get('source')
                destination = pro.get('destination')
                if show:
                    print(colored.green(" instance.name:{} provision_type:{} source:{} "
                                        "destination:{}".format(instance.name, provision_type,
                                                                source, destination)))
                else:
                    results = provision_file(vmrun, instance, source, destination)
                    LOGGER.debug('results:%s', results)
                    if results is None:
                        print(colored.red("Not Provisioned"))
                        return
                provisioned += 1

            elif provision_type == 'shell':
                inline = pro.get('inline')
                path = pro.get('path')

                args = pro.get('args')
                if not isinstance(args, list):
                    args = [args]
                if show:
                    print(colored.green(" instance.name:{} provision_type:{} inline:{} path:{} "
                                        "args:{}".format(instance.name, provision_type,
                                                         inline, path, args)))
                else:
                    if provision_shell(vmrun, instance, inline, path, args) is None:
                        print(colored.red("Not Provisioned"))
                        return
                provisioned += 1

            elif provision_type == 'pyinfra':
                path = pro.get('path')

                args = pro.get('args')
                if not isinstance(args, list):
                    args = [args]
                if show:
                    print(colored.green(" instance.name:{} provision_type:{} path:{} "
                                        "args:{}".format(instance.name, provision_type,
                                                         path, args)))
                else:
                    return_code, stdout, stderr = provision_pyinfra(instance, path, args)
                    if return_code is None:
                        print(colored.red("Not Provisioned"))
                        return
                    LOGGER.debug('return_code:%d stdout:%s stderr:%s', return_code, stdout, stderr)
                provisioned += 1

            else:
                print(colored.red("Not Provisioned ({}".format(i)))
                return
        else:
            print(colored.green("VM ({}) Provision {} "
                                "entries".format(instance.name, provisioned)))
    else:
        print(colored.blue("Nothing to provision"))


def provision_file(vmrun, instance, source, destination):
    """Provision from file.

    Args:
        vmrun (VMrun): instance of the VMrun class
        source (str): full path of a file to copy
        source (str): full path where the file is to be copied to

    Notes:
       This function copies a file from host to guest.

    """
    print(colored.blue("Copying ({}) to ({})".format(source, destination)))
    if instance.use_psk:
        results = scp(instance, source, destination, True)
    else:
        results = vmrun.copy_file_from_host_to_guest(source, destination)
    return results


def create_tempfile_in_guest(instance):
    """Create a tempfile in the guest."""
    cmd = 'tmpfile=$(mktemp); echo $tmpfile'
    _, stdout, _ = ssh(instance, cmd)
    return stdout


def provision_shell(vmrun, instance, inline, script_path, args=None):
    """Provision from shell.

    Args:
        vmrun (VMrun): instance of the VMrun class
        instance (MechInstance): instance of the MechInstance class
        inline (bool): run the script inline
        script_path (str): path to the script to run
        args (list of str): arguments to the script

    """
    if args is None:
        args = []
    if instance.use_psk:
        tmp_path = create_tempfile_in_guest(instance)
    else:
        tmp_path = vmrun.create_tempfile_in_guest()
    LOGGER.debug('inline:%s script_path:%s args:%s tmp_path:%s',
                 inline, script_path, args, tmp_path)
    if tmp_path is None or tmp_path == '':
        print(colored.red("Warning: Could not create tempfile in guest."))
        return

    try:
        if script_path and os.path.isfile(script_path):
            print(colored.blue("Configuring script {}...".format(script_path)))
            if instance.use_psk:
                results = scp(instance, script_path, tmp_path, True)
                if results is None:
                    print(colored.red("Warning: Could not copy file to guest."))
                    return
            else:
                if vmrun.copy_file_from_host_to_guest(script_path, tmp_path) is None:
                    print(colored.red("Warning: Could not copy file to guest."))
                    return
        else:
            if script_path:
                if any(script_path.startswith(s) for s in ('https://', 'http://', 'ftp://')):
                    print(colored.blue("Downloading {}...".format(script_path)))
                    try:
                        response = requests.get(script_path)
                        response.raise_for_status()
                        inline = response.read()
                    except requests.HTTPError:
                        return
                    except requests.ConnectionError:
                        return
                else:
                    print(colored.red("Cannot open {}".format(script_path)))
                    return

            if not inline:
                print(colored.red("No script to execute"))
                return

            print(colored.blue("Configuring script to run inline..."))
            the_file = tempfile.NamedTemporaryFile(delete=False)
            try:
                the_file.write(str.encode(inline))
                the_file.close()
                if instance.use_psk:
                    scp(instance, the_file.name, tmp_path, True)
                else:
                    if vmrun.copy_file_from_host_to_guest(the_file.name, tmp_path) is None:
                        return
            finally:
                os.unlink(the_file.name)

        print(colored.blue("Configuring environment..."))
        make_executable = "chmod +x '{}'".format(tmp_path)
        LOGGER.debug('make_executable:%s', make_executable)
        if instance.use_psk:
            if ssh(instance, make_executable) is None:
                print(colored.red("Warning: Could not configure script in the environment."))
                return
        else:
            if vmrun.run_script_in_guest('/bin/sh', make_executable) is None:
                print(colored.red("Warning: Could not configure script in the environment."))
                return

        print(colored.blue("Executing program..."))
        if instance.use_psk:
            args_string = ' '.join([str(elem) for elem in args])
            LOGGER.debug('args:%s args_string:%s', args, args_string)
            return ssh(instance, tmp_path, args_string)
        else:
            return vmrun.run_program_in_guest(tmp_path, args)

    finally:
        if instance.use_psk:
            return ssh(instance, 'rm -f "{}"'.format(tmp_path))
        else:
            vmrun.delete_file_in_guest(tmp_path, quiet=True)


def provision_pyinfra(instance, script_path, args=None):
    """Provision using pyinfra.

    Args:
        instance (MechInstance): instance of the MechInstance class
        script_path (str): path to the script to run, must end with .py
        args (list of str): arguments to the script

    Return:
        return_code(int): return code of the process (0=success)
        stdout(str): standard output
        stderr(str): standard error


    """
    if args is None:
        args = []

    LOGGER.debug('script_path:%s args:%s', instance, script_path, args)

    if script_path and os.path.isfile(script_path):
        return run_pyinfra_script(instance.get_ip(), instance.user,
                                  password=instance.password,
                                  script_path=script_path, args=args)
    else:
        if script_path:
            if any(script_path.startswith(s) for s in ('https://', 'http://', 'ftp://')):
                print(colored.blue("Downloading {}...".format(script_path)))
                try:
                    response = requests.get(script_path)
                    response.raise_for_status()
                    pyinfra_remote_contents = response.text
                except requests.HTTPError:
                    return
                except requests.ConnectionError:
                    return
            else:
                print(colored.red("Cannot open {}".format(script_path)))
                return

        LOGGER.debug('pyinfra_remote_contents:%s', pyinfra_remote_contents)
        the_file = tempfile.NamedTemporaryFile(delete=False, suffix='.py')
        try:
            the_file.write(str.encode(pyinfra_remote_contents))
            the_file.close()
            return run_pyinfra_script(host=instance.get_ip(), username=instance.user,
                                      password=instance.password,
                                      script_path=the_file.name, args=args)
        finally:
            os.unlink(the_file.name)


def pyinfra_installed():
    """Return True if the utility 'pyinfra' is installed, otherwise return False.
    """
    results = subprocess.run("pyinfra --version", shell=True, capture_output=True)
    return results.returncode == 0


def run_pyinfra_script(host, username, password=None, script_path=None, args=None):
    """Run a pyinfra script on host.

       Parameters
         - host(string): hostname to run the pyinfra script against, can be either
                       a hostname or ip address
         - username(string): username to use for the authentication to the host
         - password(string): password to use for the authentication to the host,
                             if no password, assume pre-shared key
         - script_path(string): the path of of the pyinfra script
         - args(kwargs): arguments to pass to running the script

       Returns:
         - return_code(int): return code from the run example: 0=success
         - stdout(string): output from the run
         - stderr(string): errors from the run

    """
    LOGGER.debug("host:%s username:%s script_path:%s args:%s",
                 host, username, script_path, args)
    if args is None:
        args = []

    if host is None or host == '':
        print(colored.red("Warning: A host is required for pyinfra provisioning."))
        return

    if username is None or username == '':
        print(colored.red("Warning: A username is required for pyinfra provisioning."))
        return

    if script_path is None or script_path == '':
        print(colored.red("Warning: A script is required for pyinfra provisioning."))
        return

    if not os.path.exists(script_path):
        print(colored.red("Warning: Could not find the pyinfra script ({}).".format(script_path)))
        return

    if not script_path.endswith('.py'):
        print(colored.red("Warning: A pyinfra provisioning script must end with .py."))
        return

    is_pyinfra_installed = pyinfra_installed()
    if not is_pyinfra_installed:
        print(colored.red("Warning: pyinfra must be installed."))
        return

    # hide password (howerver, it is probably just 'vagrant')
    password_to_print = "***using psk***"
    if password is not None:
        password_to_print = "***hidden***"
    print(colored.green("Going to run ({}) using args({}) on host:{} "
                        "authenticating with username:{} password:{}"
                        .format(script_path, args, host, username, password_to_print)))

    user_auth = '--user "{}"'.format(username)
    if password is not None:
        user_auth += ' --password "{}"'.format(username, password)
    command = "pyinfra {host} {user_auth} {script}".format(host=host, user_auth=user_auth,
                                                           script=script_path)
    LOGGER.debug('About to run this pyinfra command:%s', command)
    results = subprocess.run(command, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    return results.returncode, stdout, stderr


def config_ssh_string(config_ssh):
    """Build the ssh-config string from a dict holding the keys/values."""
    ssh_config = "Host {}".format(config_ssh.get('Host', '')) + os.linesep
    for key, value in config_ssh.items():
        if key != 'Host':
            ssh_config += "  {} {}".format(key, value) + os.linesep
    return ssh_config


def share_folders(vmrun, inst):
    """Share folders.
    Args:
        vmrun (VMrun): an instance of the VMrun class
        inst (MechInstance): an instance of the MechInstance class (representing a vm)

    """
    print(colored.blue("Sharing folders..."))
    vmrun.enable_shared_folders(quiet=False)
    for share in inst.shared_folders:
        share_name = share.get('share_name')
        host_path = share.get('host_path')
        print(colored.blue("share:{} host_path:{}".format(share_name, host_path)))
        vmrun.add_shared_folder(share_name, host_path, quiet=True)


def get_fallback_executable(command_name='vmrun'):
    """Get a fallback executable for a command line tool."""
    if 'PATH' in os.environ:
        LOGGER.debug("os.environ['PATH']:%s", os.environ['PATH'])
        for path in os.environ['PATH'].split(os.pathsep):
            vmrun = os.path.join(path, command_name)
            if os.path.exists(vmrun):
                return vmrun
            vmrun = os.path.join(path, command_name + '.exe')
            if os.path.exists(vmrun):
                return vmrun
    return None


def valid_provider(provider):
    """Determine if the provider provided is supported and valid.
       Returns True if it is valid or False if it is not valid.
    """
    if provider in ('vmware', 'virtualbox'):
        return True
    else:
        return False


def get_darwin_executable(command_name='vmrun'):
    """Get the full path for the command to run on a mac host."""
    if command_name == 'vmrun':
        full_command = '/Applications/VMware Fusion.app/Contents/Library/vmrun'
    else:
        full_command = '/usr/local/bin/VBoxManage'
    if os.path.exists(full_command):
        return full_command
    return get_fallback_executable()


def get_win32_executable():
    """Get the full path for the 'vmrun' command on a Windows host.
    """
    if PY3:
        import winreg
    else:
        import _winreg as winreg
    reg = winreg.ConnectRegistry(None, winreg.HKEY_LOCAL_MACHINE)
    try:
        key = winreg.OpenKey(reg, 'SOFTWARE\\VMware, Inc.\\VMware Workstation')
        try:
            return os.path.join(winreg.QueryValueEx(key, 'InstallPath')[0], 'vmrun.exe')
        finally:
            winreg.CloseKey(key)
    except WindowsError:
        key = winreg.OpenKey(reg, 'SOFTWARE\\WOW6432Node\\VMware, Inc.\\VMware Workstation')
        try:
            return os.path.join(winreg.QueryValueEx(key, 'InstallPath')[0], 'vmrun.exe')
        finally:
            winreg.CloseKey(key)
    finally:
        reg.Close()
    return get_fallback_executable()


def get_provider(vmrun_executable):
    """
    Identifies the right hosttype for vmrun command (ws | fusion | player)
    """

    if sys.platform == 'darwin':
        return 'fusion'

    for provider in ['ws', 'player', 'fusion']:
        # To determine the provider, try
        # running the vmrun command to see which one works.
        try:
            startupinfo = None
            if os.name == "nt":
                startupinfo = subprocess.STARTUPINFO()
                startupinfo.dwFlags |= subprocess.SW_HIDE | subprocess.STARTF_USESHOWWINDOW
            proc = subprocess.Popen([vmrun_executable,
                                     '-T',
                                     provider,
                                     'list'],
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE,
                                    startupinfo=startupinfo)
        except OSError:
            pass

        map(b2s, proc.communicate())
        if proc.returncode == 0:
            return provider



# Copyright (c) 2020 Mike Kinney

"""Mech integration tests: shared folders tests"""
import re
import subprocess

import pytest


@pytest.mark.int
def test_int_shared_folders(helpers):
    """Shared folders testing."""

    test_dir = "tests/int/shared_folders/tmp"
    helpers.cleanup_dir_and_vms_from_dir(test_dir)

    # copy file from parent dir
    command = "cp ../Mechfile ."
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stdout == ''
    assert stderr == ''
    assert results.returncode == 0

    # up
    command = "mech up"
    expected_lines = [r"started"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # ensure the Mechfile is present from the guest
    command = 'mech ssh -c "ls -al /mnt/hgfs/mech/Mechfile" first'
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert re.search("Mechfile", stdout)
    assert results.returncode == 0

    # create a simple file to see if visible on another share
    command = 'date > /tmp/now'
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stdout == ''
    assert stderr == ''
    assert results.returncode == 0

    # ensure we can see the file on the other share
    command = 'mech ssh -c "ls -al /mnt/hgfs/mech2/now" first'
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert re.search("/mnt/hgfs/mech2/now", stdout)
    assert results.returncode == 0

    # stop
    command = "mech stop"
    expected_lines = [r"Stopped"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # "up" but do not have shared folders
    command = "mech up --disable-shared-folders"
    expected_lines = [r"started"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # ensure the Mechfile is *NOT* present from the guest
    command = 'mech ssh -c "ls -al /mnt/hgfs/mech/Mechfile" first'
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    # assert re.search("No such file or directory", stdout)
    # linux returns 255 and nothing on stdout
    assert results.returncode == 2 or results.returncode == 255

    # pause
    command = "mech pause"
    expected_lines = [r"Paused"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # "resume" but do not have shared folders
    command = "mech resume --disable-shared-folders"
    expected_lines = ["resumed"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # ensure the Mechfile is *NOT* present from the guest
    command = 'mech ssh -c "ls -al /mnt/hgfs/mech/Mechfile" first'
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert re.search("No such file or directory", stdout)
    assert results.returncode == 2

    # destroy
    command = "mech destroy -f"
    expected = "Deleting"
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    assert re.search(expected, stdout)

    # clean up at the end
    helpers.cleanup_dir_and_vms_from_dir(test_dir)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016-2017 Kevin Chung
# Copyright (c) 2018 German Mendez Bravo (Kronuz)
# Copyright (c) 2020 Mike Kinney
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
"""MechBox class"""

from __future__ import print_function, absolute_import

import os
import sys
import fnmatch
import logging
import shutil


from clint.textui import colored


from . import utils
from .mech_command import MechCommand

LOGGER = logging.getLogger(__name__)


class MechBox(MechCommand):
    """
    Usage: mech box <subcommand> [<args>...]

    Available subcommands:
        add               add a box to the catalog of available boxes
        (list|ls)         list available boxes in the catalog
        (remove|delete)   removes a box that matches the given name

    For help on any individual subcommand run `mech box <subcommand> -h`
    """

    def add(self, arguments):  # pylint: disable=no-self-use
        """
        Add a box to the catalog of available boxes.

        Usage: mech box add [options] <location>

        Notes:
            The location can be a:
                URL (ex: 'http://example.com/foo.box'),
                box file (ex: 'file:/mnt/boxen/foo.box'),
                json file (ex: 'file:/tmp/foo.json'), or
                HashiCorp account/box (ex: 'bento/ubuntu-18.04').

        Options:
                --box-version VERSION        Constrain version of the added box
            -f, --force                      Overwrite an existing box if it exists
            -h, --help                       Print this help
            -p, --provider PROVIDER          Provider ('vmware' or 'virtualbox')
        """

        location = arguments['<location>']
        box_version = arguments['--box-version']
        force = arguments['--force']
        provider = arguments['<provider>']

        if provider is None:
            provider = 'vmware'
        if not utils.valid_provider(provider):
            sys.exit(colored.red("Need to provide valid provider."))

        utils.add_box(name=None, box=None, location=location, box_version=box_version,
                      force=force, provider=provider)

    def list(self, arguments):  # pylint: disable=no-self-use,unused-argument
        """
        List all available boxes in the catalog.

        Usage: mech box list [options]

        Options:
            -h, --help                       Print this help
        """

        print("{}\t{}\t{}".format(
            'BOX'.rjust(35),
            'VERSION'.rjust(12),
            'PROVIDER'.rjust(12),
        ))
        path = os.path.abspath(os.path.join(utils.mech_dir(), 'boxes'))
        for root, _, filenames in os.walk(path):
            for filename in fnmatch.filter(filenames, '*.box'):
                directory = os.path.dirname(os.path.join(root, filename))[len(path) + 1:]
                provider, account, box, version = directory.split('/', 3)
                print("{}\t{}\t{}".format(
                    "{}/{}".format(account, box).rjust(35),
                    version.rjust(12),
                    provider.rjust(12),
                ))

    # add alias for 'mech box ls'
    ls = list

    def remove(self, arguments):  # pylint: disable=no-self-use
        """
        Remove a box from mech that matches the given name and version.

        Usage: mech box remove [options] <provider> <name> <version>

        Options:
            -h, --help                       Print this help
        """
        name = arguments['<name>']
        box_version = arguments['<version>']
        provider = arguments['<provider>']

        if not utils.valid_provider(provider):
            sys.exit(colored.red("Need to provide valid provider."))

        path = os.path.abspath(os.path.join(utils.mech_dir(), 'boxes', provider, name, box_version))
        if os.path.exists(path):
            shutil.rmtree(path)
            print("Removed {} {}".format(name, box_version))
        else:
            print("No boxes were removed.")

    # add alias for 'mech box delete'
    delete = remove



# Copyright (c) 2020 Mike Kinney

"""Mech integration tests: add and remove instances"""
import re
import subprocess


import pytest


@pytest.mark.int
def test_int_add_and_remove_instances_using_add_first(helpers):
    """Test adding/removing of instances from Mechfile starting off with add."""

    test_dir = "tests/int/add_and_remove_instances_add_first"
    helpers.cleanup_dir_and_vms_from_dir(test_dir)

    command = "mech add apple bento/ubuntu-18.04"
    expected_lines = ["Adding", "Added"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    command = "mech list"
    expected_lines = ["apple"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    command = "mech add banana bento/ubuntu-18.04"
    expected_lines = ["Adding", "Added"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    command = "mech list"
    expected_lines = ["apple", "banana"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    command = "mech remove banana"
    expected_lines = ["Removed"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    command = "mech list"
    expected_lines = ["apple"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # should be able to destroy
    command = "mech destroy -f"
    expected = "not created"
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    assert re.search(expected, stdout)


@pytest.mark.int
def test_int_add_and_remove_instances_using_init_first(helpers):
    """Test adding/removing of instances from Mechfile starting off with init."""

    test_dir = "tests/int/add_and_remove_instances_init_first"
    helpers.cleanup_dir_and_vms_from_dir(test_dir)

    command = "mech init --name apple bento/ubuntu-18.04"
    expected_lines = ["Initializing", "has been init"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    command = "mech list"
    expected_lines = ["apple"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    command = "mech add banana bento/ubuntu-18.04"
    expected_lines = ["Adding", "Added"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    command = "mech list"
    expected_lines = ["apple", "banana"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    command = "mech remove banana"
    expected_lines = ["Removing", "Removed"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    command = "mech list"
    expected_lines = ["apple"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # should be able to destroy
    command = "mech destroy -f"
    expected = "not created"
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    assert re.search(expected, stdout)



# Copyright (c) 2020 Mike Kinney

"""Mech integration tests: simple ones (including virtualbox smoke)"""
import re
import subprocess
import platform

import pytest


@pytest.mark.int
def test_int_smoke_virtualbox(helpers):
    """Smoke test most options using virtualbox."""

    test_dir = "tests/int/simple_virtualbox"
    helpers.cleanup_dir_and_vms_from_dir(test_dir, 'first')

    # should init
    command = "mech init --provider virtualbox bento/ubuntu-18.04"
    expected_lines = ["Initializing", "Loading metadata", "has been initialized", "mech up"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # should start
    command = "mech up"
    expected_lines = ["virtualbox", "Extracting", "Getting IP", "started",
                      "Not yet implemented"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # should be able to re-up, verify 'start' alias works, too
    commands = ["mech up", "mech start"]
    expected_lines = ["Getting IP", "started"]
    for command in commands:
        results = subprocess.run(commands, cwd=test_dir, shell=True, capture_output=True)
        stdout = results.stdout.decode('utf-8')
        stderr = results.stderr.decode('utf-8')
        assert stderr == ''
        assert results.returncode == 0
        for line in expected_lines:
            print(line)
            assert re.search(line, stdout, re.MULTILINE)

    # test 'mech ps'
    command = "mech ps first"
    expected_lines = ["Not yet implemented"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # test 'mech status'
    command = "mech status"
    expected_lines = ["Not yet implemented"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # test 'mech global-status'
    command = "mech global-status"
    expected_lines = ['first']
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # test 'mech list'
    commands = ["mech ls", "mech list"]
    expected_lines = ['first', 'ubuntu', 'virtualbox']
    for command in commands:
        results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
        stdout = results.stdout.decode('utf-8')
        stderr = results.stderr.decode('utf-8')
        assert stderr == ''
        assert results.returncode == 0
        for line in expected_lines:
            print(line)
            assert re.search(line, stdout, re.MULTILINE)

    # test 'mech stop'
    command = "mech stop"
    expected_lines = ['Stopped']
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # test 'mech stop' again
    command = "mech stop"
    expected_lines = ['Not stopped']
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # test 'mech start'
    command = "mech start"
    expected_lines = ['started']
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # test 'mech pause'
    command = "mech pause"
    expected_lines = ['Not yet implemented']
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # test 'mech resume'
    command = "mech resume"
    expected_lines = ['Not yet implemented']
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # test 'mech suspend'
    command = "mech suspend"
    expected_lines = ['Not yet implemented']
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # test 'mech resume' after suspend
    command = "mech resume"
    expected_lines = ['Not yet implemented']
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # test 'mech ssh' (different forms)
    commands = ["mech ssh -c 'uptime' first", "mech ssh --command 'uptime' first"]
    expected_lines = ['load average']
    for command in commands:
        results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
        stdout = results.stdout.decode('utf-8')
        stderr = results.stderr.decode('utf-8')
        assert stderr == ''
        assert results.returncode == 0
        for line in expected_lines:
            print(line)
            assert re.search(line, stdout, re.MULTILINE)

    # test 'mech scp' to guest
    command = "date > now; mech scp now first:/tmp"
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stdout == ''
    assert stderr == ''
    assert results.returncode == 0

    # test 'mech scp' from guest
    command = "mech scp first:/tmp/now ."
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stdout == ''
    assert stderr == ''
    assert results.returncode == 0

    # test 'mech ip first'
    command = "mech ip first"
    expected = r"[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}"
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    assert re.search(expected, stdout)

    if platform.system() == "Linux":
        # test "mech port"
        command = "mech port"
        results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
        stdout = results.stdout.decode('utf-8')
        stderr = results.stderr.decode('utf-8')
        assert stdout == ''
        assert stderr.strip() == 'This command is not supported on this OS.'
        assert results.returncode == 1
    else:
        # test "mech port"
        command = "mech port"
        expected = "Not yet implemented"
        results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
        stdout = results.stdout.decode('utf-8')
        stderr = results.stderr.decode('utf-8')
        assert stderr == ''
        assert results.returncode == 0
        assert re.search(expected, stdout, re.MULTILINE)

    # test "mech box list" (and alias)
    commands = ["mech box list", "mech box ls"]
    expected = r"ubuntu"
    for command in commands:
        results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
        stdout = results.stdout.decode('utf-8')
        stderr = results.stderr.decode('utf-8')
        assert stderr == ''
        assert results.returncode == 0
        assert re.search(expected, stdout, re.MULTILINE)

    # test "mech snapshot list" (and alias)
    commands = ["mech snapshot list", "mech snapshot ls"]
    expected = "Not yet implemented"
    for command in commands:
        results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
        stdout = results.stdout.decode('utf-8')
        stderr = results.stderr.decode('utf-8')
        assert stderr == ''
        assert results.returncode == 0
        assert re.search(expected, stdout, re.MULTILINE)

    # test "mech snapshot save"
    command = "mech snapshot save snap1 first"
    expected = "Not yet implemented"
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    assert re.search(expected, stdout)

    # test "mech snapshot save" with same args again
    # command = "mech snapshot save snap1 first"
    # results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    # stdout = results.stdout.decode('utf-8')
    # stderr = results.stderr.decode('utf-8')
    # assert stdout == ''
    # assert re.search('A snapshot with the name already exists', stderr)
    # assert results.returncode == 1

    # test "mech snapshot list" (and alias) again (now that we have one)
    # commands = ["mech snapshot list", "mech snapshot ls"]
    # expected = "Total snapshots: 1"
    # for command in commands:
    #     results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    #     stdout = results.stdout.decode('utf-8')
    #     stderr = results.stderr.decode('utf-8')
    #     assert stderr == ''
    #     assert results.returncode == 0
    #     assert re.search(expected, stdout, re.MULTILINE)

    # test "mech snapshot delete"
    command = "mech snapshot delete snap1 first"
    expected = "Not yet implemented"
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    assert re.search(expected, stdout)

    # should be able to destroy
    command = "mech destroy -f"
    expected = "Deleting"
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    assert re.search(expected, stdout)

    # clean up at the end
    helpers.cleanup_dir_and_vms_from_dir(test_dir)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016-2017 Kevin Chung
# Copyright (c) 2018 German Mendez Bravo (Kronuz)
# Copyright (c) 2020 Mike Kinney
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
"""Mech class"""

from __future__ import print_function, absolute_import

import os
import platform
import sys
import logging
import textwrap
import shutil

from clint.textui import colored

from . import utils
from .vmrun import VMrun
from .vbm import VBoxManage
from .mech_instance import MechInstance
from .mech_command import MechCommand
from .mech_box import MechBox
from .mech_snapshot import MechSnapshot

LOGGER = logging.getLogger(__name__)


class Mech(MechCommand):
    """
    Usage: mech [options] <command> [<args>...]

    Options:
        -v, --version                    Print the version and exit.
        -h, --help                       Print this help.
        --debug                          Show debug messages.

    Common commands:
        box               manages boxes: add, list remove, etc.
        destroy           stops and deletes all traces of the instances
        (down|stop|halt)  stops the instances
        global-status     outputs status of all virutal machines on this host
        init              initializes a new Mech environment by creating a Mechfile
        ip                outputs ip of an instance
        (list|ls)         lists all available boxes
        pause             pauses the instances
        port              displays information about guest port mappings
        provision         provisions the Mech machine
        ps                list running processes for an instance
        reload            restarts Mech machine, loads new Mechfile configuration
        resume            resume a paused/suspended Mech machine
        scp               copies files to/from the machine via SCP
        snapshot          manages snapshots: save, list, remove, etc.
        ssh               connects to an instance via SSH
        ssh-config        outputs OpenSSH valid configuration to connect to the instances
        status            outputs status of the instances
        suspend           suspends the instances
        (up|start)        starts instances (aka virtual machines)
        upgrade           upgrade the instances

    For help on any individual command run `mech <command> -h`

    All "state" will be saved in .mech directory. (boxes and instances)

    Example:

    Initializing and using a machine from HashiCorp's Vagrant Cloud:

        mech init bento/ubuntu-18.04
        mech up
        mech ssh
    """

    subcommand_name = '<command>'

    def __init__(self, arguments):
        super(Mech, self).__init__(arguments)

        logger = logging.getLogger()
        handler = logging.StreamHandler(sys.stderr)
        formatter = logging.Formatter('%(filename)s:%(lineno)s %(funcName)s() '
                                      '%(levelname)s: %(message)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        if arguments['--debug']:
            logger.setLevel(logging.DEBUG)

    box = MechBox
    snapshot = MechSnapshot

    def init(self, arguments):  # pylint: disable=no-self-use
        """
        Initializes a new mech environment by creating a Mechfile.

        Usage: mech init [options] <location>

        Notes:
          - The location can be a:
              + URL (ex: 'http://example.com/foo.box'),
              + box file (ex: 'file:/mnt/boxen/foo.box'),
              + json file (ex: 'file:/tmp/foo.json'), or
              + HashiCorp account/box (ex: 'bento/ubuntu-18.04').
          - A default shared folder name 'mech' will be available
            in the guest for the current directory.
          - The 'add-me' option will add the currently logged in user to the guest,
            add the same user to sudoers, and add the id_rsa.pub key to the
            authorized_hosts file for that user.
          - The 'use-me' option will use the currently logged in user for
            future interactions with the guest instead of the vagrant user.
            The first provisioning run will be done with 'vagrant' user.

        Options:
            -a, --add-me                     Add the current host user/pubkey to guest
                --box BOXNAME                Name of the box (ex: bento/ubuntu-18.04)
                --box-version VERSION        Constrain version of the added box
            -f, --force                      Overwrite existing Mechfile
            -h, --help                       Print this help
                --name INSTANCE              Name of the instance (myinst1)
            -p, --provider PROVIDER          Provider ('vmware' or 'virtualbox')
            -u, --use-me                     Use the current user for mech interactions
        """
        add_me = arguments['--add-me']
        use_me = arguments['--use-me']
        name = arguments['--name']
        box_version = arguments['--box-version']
        box = arguments['--box']
        location = arguments['<location>']
        provider = arguments['--provider']

        if provider is None:
            provider = 'vmware'
        if not utils.valid_provider(provider):
            sys.exit(colored.red("Need to provide valid provider."))

        if not name or name == "":
            name = "first"

        force = arguments['--force']

        LOGGER.debug('name:%s box:%s box_version:%s location:%s provider:%s',
                     name, box, box_version, location, provider)

        if os.path.exists('Mechfile') and not force:
            sys.exit(colored.red(textwrap.fill(
                "`Mechfile` already exists in this directory. Remove it "
                "before running `mech init`.")))

        print(colored.green("Initializing mech"))
        utils.init_mechfile(
            location=location,
            box=box,
            name=name,
            box_version=box_version,
            add_me=add_me,
            use_me=use_me,
            provider=provider)
        print(colored.green(textwrap.fill(
            "A `Mechfile` has been initialized and placed in this directory. "
            "You are now ready to `mech up` your first virtual environment!")))

    def add(self, arguments):  # pylint: disable=no-self-use
        """
        Add instance to the Mechfile.

        Usage: mech add [options] <name> <location>

        Example box: bento/ubuntu-18.04

        Notes:
        - The 'add-me' option will add the currently logged in user to the guest,
          add the same user to sudoers, and add the id_rsa.pub key to the authorized_hosts file
          for that user.

        Options:
            -a, --add-me                     Add the current host user/pubkey to guest
                --box BOXNAME                Name of the box (ex: bento/ubuntu-18.04)
                --box-version VERSION        Constrain version of the added box
            -h, --help                       Print this help
            -p, --provider PROVIDER          Provider ('vmware' or 'virtualbox')
            -u, --use-me                     Use the current user for mech interactions
        """
        name = arguments['<name>']
        box_version = arguments['--box-version']
        box = arguments['--box']
        add_me = arguments['--add-me']
        use_me = arguments['--use-me']
        provider = arguments['--provider']
        location = arguments['<location>']

        print('before provider:{}'.format(provider))
        if provider is None:
            provider = 'vmware'
        if not utils.valid_provider(provider):
            sys.exit(colored.red("Need to provide valid provider."))
        print('after provider:{}'.format(provider))

        if not name or name == "":
            sys.exit(colored.red("Need to provide a name for the instance to add to the Mechfile."))

        LOGGER.debug('name:%s box:%s box_version:%s location:%s provider:%s',
                     name, box, box_version, location, provider)

        print(colored.green("Adding ({}) to the Mechfile.".format(name)))

        utils.add_to_mechfile(
            location=location,
            box=box,
            name=name,
            box_version=box_version,
            add_me=add_me,
            use_me=use_me,
            provider=provider)
        print(colored.green("Added to the Mechfile."))

    def remove(self, arguments):
        """
        Remove instance from the Mechfile.

        Usage: mech remove [options] <name>

        Options:
            -h, --help                       Print this help
        """
        name = arguments['<name>']

        if not name or name == "":
            sys.exit(colored.red("Need to provide a name to be removed from the Mechfile."))

        LOGGER.debug('name:%s', name)

        self.activate_mechfile()
        inst = self.mechfile.get(name, None)
        if inst:
            print(colored.green("Removing ({}) from the Mechfile.".format(name)))
            utils.remove_mechfile_entry(name=name)
            print(colored.green("Removed from the Mechfile."))
        else:
            sys.exit(colored.red("There is no instance called ({}) in the Mechfile.".format(name)))

    # add aliases for 'mech delete'
    delete = remove
    rm = remove

    def up(self, arguments):  # pylint: disable=invalid-name
        """
        Starts and provisions the mech environment.

        Usage: mech up [options] [<instance>]

        Notes:
           - If no instance is specified, all instances will be started.
           - The options (memsize, numvcpus, and no-nat) will only be applied
             upon first run of the 'up' command.
           - The 'no-nat' option will only be applied if there is no network
             interface supplied in the box file.
           - Unless 'disable-shared-folders' is used, a default read/write
             share called "mech" will be mounted from the current directory.
             (ex: '/mnt/hgfs/mech' on guest will have the file "Mechfile".)
             To change shared folders, modify the Mechfile directly.
           - The 'remove-vagrant' option will remove the vagrant account from the
             guest VM which is what 'mech' uses to communicate with the VM.
             Be sure you can connect/admin the instance before using this option.

        Options:
                --disable-provisioning       Do not provision
                --disable-shared-folders     Do not share folders with VM
                --gui                        Start GUI
                --memsize 1024               Specify the size of memory for VM
                --no-cache                   Do not save the downloaded box
                --no-nat                     Do not use NAT network (i.e., bridged)
                --numvcpus 1                 Specify the number of vcpus for VM
            -h, --help                       Print this help
            -r, --remove-vagrant             Remove vagrant user
        """
        gui = arguments['--gui']
        disable_shared_folders = arguments['--disable-shared-folders']
        disable_provisioning = arguments['--disable-provisioning']
        save = not arguments['--no-cache']
        remove_vagrant = arguments['--remove-vagrant']

        memsize = arguments['--memsize']
        numvcpus = arguments['--numvcpus']
        no_nat = arguments['--no-nat']

        instance_name = arguments['<instance>']

        LOGGER.debug('gui:%s disable_shared_folders:%s disable_provisioning:%s '
                     'save:%s numvcpus:%s memsize:%s no_nat:%s', gui,
                     disable_shared_folders, disable_provisioning, save,
                     numvcpus, memsize, no_nat)

        if instance_name:
            # single instance
            instances = [instance_name]
        else:
            # multiple instances
            instances = self.instances()

        for instance in instances:
            inst = MechInstance(instance)

            location = inst.url
            if not location:
                location = inst.box_file

            # only run init_box on first "up"
            # extracts the VM files from the singular .box archive
            if not inst.created:
                path_to_vmx_or_ovf = utils.init_box(
                    instance,
                    box=inst.box,
                    box_version=inst.box_version,
                    location=location,
                    instance_path=inst.path,
                    save=save,
                    numvcpus=numvcpus,
                    memsize=memsize,
                    no_nat=no_nat,
                    provider=inst.provider)
                if inst.provider == 'vmware':
                    inst.vmx = path_to_vmx_or_ovf
                else:
                    inst.ovf = path_to_vmx_or_ovf
                inst.created = True

            started = None
            if inst.provider == 'vmware':
                # Note: user/password is needed for provisioning
                vmrun = VMrun(inst.vmx, user=inst.user, password=inst.password)
                print(colored.blue("Bringing machine ({}) up...".format(instance)))
                started = vmrun.start(gui=gui)
            else:
                vbm = VBoxManage()
                vbm.create_hostonly(quiet=True)
                vbm.hostonly(inst.name, quiet=True)
                vbm.start(vmname=inst.name, gui=gui, quiet=True)
                running_vms = vbm.list_running()
                started = None
                if inst.name in running_vms:
                    started = True

            if started is None:
                print(colored.red("VM not started"))
            else:
                print(colored.blue("Getting IP address..."))
                ip_address = inst.get_ip(wait=True)

                if not disable_shared_folders:
                    # TODO: virtualbox
                    if inst.provider == 'vmware':
                        utils.share_folders(vmrun, inst)

                if ip_address:
                    print(colored.green("VM ({})started on {}".format(instance, ip_address)))
                else:
                    print(colored.green("VM ({}) started on an unknown "
                                        "IP address".format(instance)))

                if inst.provider == 'vmware':
                    # if not already using preshared key, switch to it
                    if not inst.use_psk and inst.auth:
                        utils.add_auth(inst)
                        inst.switch_to_psk()

                    if remove_vagrant:
                        utils.del_user(inst, 'vagrant')

                    if not disable_provisioning:
                        utils.provision(inst, show=False)
                else:
                    print(colored.red("Not yet implemented on this platform:"))
                    print(colored.red("(use-psk, remove-vagrant, provisioning)"))

    # allows "mech start" to alias to "mech up"
    start = up

    def global_status(self, arguments):  # pylint: disable=no-self-use,unused-argument
        """
        Outputs info about all VMs running on this computer.

        Usage: mech global-status [options]

        Options:
            -h, --help                       Print this help
        """
        vmrun = VMrun()
        if vmrun.installed():
            print("===VMware VMs===")
            print(vmrun.list())

        vbm = VBoxManage()
        if vbm.installed():
            print("===VirtualBox VMs===")
            print(vbm.list())

    def ps(self, arguments):  # pylint: disable=invalid-name,no-self-use
        """
        List running processes in Guest OS.

        Usage: mech ps [options] <instance>

        Options:
            -h, --help                       Print this help
        """
        instance_name = arguments['<instance>']

        inst = MechInstance(instance_name)

        if inst.created:
            if inst.provider == 'vmware':
                # Note: user/password is needed for ps
                vmrun = VMrun(inst.vmx, inst.user, inst.password)
                print(vmrun.list_processes_in_guest())
            else:
                print(colored.red("Not yet implemented on this platform."))

        else:
            print("VM {} not created.".format(instance_name))

    # alias "mech process_status" to "mech ps"
    process_status = ps

    def status(self, arguments):
        """
        Outputs status of the instances.

        Usage: mech status [options] [<instance>]

        Options:
            -h, --help                       Print this help
        """
        instance_name = arguments['<instance>']

        if instance_name:
            # single instance
            instances = [instance_name]
        else:
            # multiple instances
            instances = self.instances()

        for instance in instances:
            inst = MechInstance(instance)

            if inst.created:

                if inst.provider == 'vmware':
                    vmrun = VMrun(inst.vmx)

                    ip_address = inst.get_ip()
                    state = vmrun.check_tools_state(quiet=True)

                    print("Current machine state:" + os.linesep)
                    if ip_address is None:
                        ip_address = "poweroff"
                    elif not ip_address:
                        ip_address = "unknown"
                    print("%s\t%s\t%s\t(VMware Tools %s)" % (inst.name, inst.box,
                                                             ip_address, state))

                    if ip_address == "poweroff":
                        print(os.linesep + "The VM is powered off. To restart the VM, "
                              "simply run `mech up {}`".format(instance))
                    elif ip_address == "unknown":
                        print(os.linesep + "The VM is on. but it has no IP to connect to,"
                              "VMware Tools must be installed")
                    elif state in ("installed", "running"):
                        print(os.linesep + "The VM is ready. Connect to it "
                              "using `mech ssh {}`".format(instance))
                else:
                    print(colored.red("Not yet implemented on this platform."))
            else:
                print("The VM ({}) has not been created.".format(instance))

    def destroy(self, arguments):
        """
        Stops and deletes all traces of the instances.

        Usage: mech destroy [options] [<instance>]

        Options:
            -f, --force                      Destroy without confirmation.
            -h, --help                       Print this help
        """
        force = arguments['--force']

        instance_name = arguments['<instance>']

        if instance_name:
            # single instance
            instances = [instance_name]
        else:
            # multiple instances
            instances = self.instances()

        for instance in instances:
            inst = MechInstance(instance)

            if os.path.exists(inst.path):
                if force or utils.confirm("Are you sure you want to delete {} "
                                          "at {}".format(inst.name, inst.path), default='n'):
                    print(colored.green("Deleting ({})...".format(instance)))

                    if inst.provider == 'vmware':
                        vmrun = VMrun(inst.vmx)
                        vmrun.stop(mode='hard', quiet=True)
                        vmrun.delete_vm()
                    else:
                        vbm = VBoxManage()

                        vbm.stop(vmname=inst.name, quiet=True)
                        vbm.unregister(vmname=inst.name, quiet=True)

                    if os.path.exists(inst.path):
                        shutil.rmtree(inst.path)
                    print("Deleted")
                else:
                    print(colored.red("Delete aborted."))
            else:
                print(colored.red("VM ({}) not created.".format(instance)))

    def down(self, arguments):
        """
        Stops the instances.

        Usage: mech down [options] [<instance>]

        Options:
            -f, --force                      Force a hard stop
            -h, --help                       Print this help
        """
        force = arguments['--force']

        instance_name = arguments['<instance>']

        if instance_name:
            # single instance
            instances = [instance_name]
        else:
            # multiple instances
            instances = self.instances()

        for instance in instances:
            inst = MechInstance(instance)

            if inst.created:
                if inst.provider == 'vmware':
                    vmrun = VMrun(inst.vmx)
                    if not force and vmrun.installed_tools():
                        stopped = vmrun.stop()
                    else:
                        stopped = vmrun.stop(mode='hard')
                    if stopped is None:
                        print(colored.red("Not stopped", vmrun))
                    else:
                        print(colored.green("Stopped", vmrun))
                else:
                    vbm = VBoxManage()
                    stopped = vbm.stop(vmname=inst.name, quiet=True)
                    if stopped is None:
                        print(colored.red("Not stopped", vbm))
                    else:
                        print(colored.green("Stopped", vbm))
            else:
                print(colored.red("VM ({}) not created.".format(instance)))

    # alias 'mech stop' and 'mech halt' to 'mech down'
    stop = down
    halt = down

    def pause(self, arguments):
        """
        Pauses the instances.

        Usage: mech pause [options] [<instance>]

        Options:
            -h, --help                       Print this help
        """
        instance_name = arguments['<instance>']

        if instance_name:
            # single instance
            instances = [instance_name]
        else:
            # multiple instances
            instances = self.instances()

        for instance in instances:
            inst = MechInstance(instance)

            if inst.created:
                if inst.provider == 'vmware':
                    vmrun = VMrun(inst.vmx)
                    pause_results = vmrun.pause()
                    if pause_results is None:
                        print(colored.red("Not paused", vmrun))
                    else:
                        print(colored.yellow("Paused", vmrun))
                else:
                    print(colored.red("Not yet implemented on this platform."))
            else:
                print(colored.red("VM ({}) not created.".format(instance)))

    def upgrade(self, arguments):
        """
        Upgrade the vm file format and virtual hardware for the instance(s).

        Usage: mech upgrade [options] [<instance>]

        Note: The VMs must be created and stopped.

        Options:
            -h, --help                       Print this help
        """
        instance_name = arguments['<instance>']

        if instance_name:
            # single instance
            instances = [instance_name]
        else:
            # multiple instances
            instances = self.instances()

        for instance in instances:
            inst = MechInstance(instance)

            if inst.created:
                if inst.provider == 'vmware':
                    vmrun = VMrun(inst.vmx)
                    state = vmrun.check_tools_state(quiet=True)
                    if state == "running":
                        print("VM must be stopped before doing upgrade.")
                    else:
                        if vmrun.upgradevm(quiet=False) is None:
                            print(colored.red("Not upgraded", vmrun))
                        else:
                            print(colored.yellow("Upgraded", vmrun))
                else:
                    print(colored.red("Not yet implemented on this platform."))
            else:
                print(colored.red("VM ({}) not created.".format(instance)))

    def resume(self, arguments):
        """
        Resume a paused/suspended instances.

        Usage: mech resume [options] [<instance>]

        Options:
            --disable-shared-folders         Do not share folders with VM
            -h, --help                       Print this help
        """
        instance_name = arguments['<instance>']
        disable_shared_folders = arguments['--disable-shared-folders']

        LOGGER.debug('instance_name:%s '
                     'disable_shared_folders:%s', instance_name, disable_shared_folders)

        if instance_name:
            # single instance
            instances = [instance_name]
        else:
            # multiple instances
            instances = self.instances()

        for instance in instances:
            inst = MechInstance(instance)
            LOGGER.debug('instance:%s inst.vmx:%s', instance, inst.vmx)

            # if we have started this instance before, try to unpause
            if inst.created:

                if inst.provider == 'vmware':
                    vmrun = VMrun(inst.vmx)

                    if vmrun.unpause(quiet=True) is not None:
                        print(colored.blue("Getting IP address..."))
                        ip_address = inst.get_ip(wait=True)
                        if not disable_shared_folders:
                            utils.share_folders(vmrun, inst)
                        else:
                            print(colored.blue("Disabling shared folders..."))
                            vmrun.disable_shared_folders(quiet=False)
                        if ip_address:
                            print(colored.green("VM resumed on {}".format(ip_address)))
                        else:
                            print(colored.green("VM resumed on an unknown IP address"))

                    else:
                        # Otherwise try starting
                        vmrun = VMrun(inst.vmx)
                        started = vmrun.start()
                        if started is None:
                            print(colored.red("VM not started"))
                        else:
                            print(colored.blue("Getting IP address..."))
                            ip_address = inst.get_ip(wait=True)
                            if not disable_shared_folders:
                                utils.share_folders(vmrun, inst)
                            if ip_address:
                                if started:
                                    print(colored.green("VM ({}) started on "
                                                        "{}".format(instance, ip_address)))
                                else:
                                    print(colored.yellow("VM ({}) already was started "
                                                         "on {}".format(instance, ip_address)))
                            else:
                                if started:
                                    print(colored.green("VM ({}) started on an unknown "
                                                        "IP address".format(instance)))
                                else:
                                    print(colored.yellow("VM ({}) already was started on an "
                                                         "unknown IP address".format(instance)))
                else:
                    print(colored.red("Not yet implemented on this platform."))
            else:
                print(colored.red("VM not created"))

    def suspend(self, arguments):
        """
        Suspends instances.

        Usage: mech suspend [options] [<instance>]

        Options:
            -h, --help                       Print this help
        """
        instance_name = arguments['<instance>']

        if instance_name:
            # single instance
            instances = [instance_name]
        else:
            # multiple instances
            instances = self.instances()

        for instance in instances:
            inst = MechInstance(instance)

            if inst.created:
                if inst.provider == 'vmware':
                    vmrun = VMrun(inst.vmx)
                    if vmrun.suspend() is None:
                        print(colored.red("Not suspended", vmrun))
                    else:
                        print(colored.green("Suspended", vmrun))
                else:
                    print(colored.red("Not yet implemented on this platform."))
            else:
                print("VM has not been created.")

    def ssh_config(self, arguments):
        """
        Output OpenSSH valid configuration to connect to the machine.

        Usage: mech ssh-config [options] [<instance>]

        Options:
            -h, --help                       Print this help
        """
        instance_name = arguments['<instance>']

        if instance_name:
            # single instance
            instances = [instance_name]
        else:
            # multiple instances
            instances = self.instances()

        for instance in instances:
            inst = MechInstance(instance)
            if inst.created:
                print(utils.config_ssh_string(inst.config_ssh()))
            else:
                print(colored.red("VM ({}) is not created.".format(instance)))

    def ssh(self, arguments):  # pylint: disable=no-self-use
        """
        Connects to machine via SSH.

        Usage: mech ssh [options] <instance> [-- <extra-ssh-args>...]

        Options:
            -c, --command COMMAND            Execute an SSH command directly
            -p, --plain                      Plain mode, leaves authentication up to user
            -h, --help                       Print this help
        """
        plain = arguments['--plain']
        extra = arguments['<extra-ssh-args>']
        command = arguments['--command']

        instance = arguments['<instance>']

        inst = MechInstance(instance)

        if inst.created:
            rc, stdout, stderr = utils.ssh(inst, command, plain, extra)
            LOGGER.debug('command:%s rc:%d stdout:%s stderr:%s', command, rc, stdout, stderr)
            if stdout:
                print(stdout)
            if stderr:
                print(stderr)
            sys.exit(rc)
        else:
            print("VM not created.")

    def scp(self, arguments):  # pylint: disable=no-self-use
        """
        Copies files to and from the machine via SCP.

        Usage: mech scp [options] <src> <dst> [-- <extra-ssh-args>...]

        Options:
            -h, --help                       Print this help
        """
        extra = arguments['<extra-ssh-args>']
        src = arguments['<src>']
        dst = arguments['<dst>']

        dst_instance, dst_is_host, dst = dst.partition(':')
        src_instance, src_is_host, src = src.partition(':')

        instance_name = None
        if dst_is_host and src_is_host:
            sys.exit(colored.red("Both src and dst are host destinations"))
        if dst_is_host:
            instance_name = dst_instance
        else:
            dst = dst_instance
        if src_is_host:
            instance_name = src_instance
        else:
            src = src_instance

        if instance_name is None:
            sys.exit(colored.red("Could not determine instance name."))

        inst = MechInstance(instance_name)

        if inst.created:
            utils.scp(inst, src, dst, dst_is_host, extra)
        else:
            print(colored.red('VM not created.'))

    def ip(self, arguments):  # pylint: disable=invalid-name,no-self-use
        """
        Outputs ip of the Mech machine.

        Usage: mech ip [options] <instance>

        Options:
            -h, --help                       Print this help
        """
        instance = arguments['<instance>']

        inst = MechInstance(instance)

        if inst.created:
            ip_address = inst.get_ip()
            if ip_address:
                print(colored.green(ip_address))
            else:
                print(colored.red("Unknown IP address"))
        else:
            print(colored.yellow("VM not created"))

    # alias 'mech ip_address' to 'mech ip'
    ip_address = ip

    def provision(self, arguments):
        """
        Provisions the Mech machine.

        Usage: mech provision [options] [<instance>]

        Notes:
          - There are a few provision types: 'file', 'shell', and 'pyinfra'.
          - 'shell' can be inline.
          - 'shell' and 'pyinfra' can have a remote endpoint ('http', 'https',
            'ftp') for the script.
            (ex: 'http://example.com/somefile.sh' or ex: 'ftp://foo.com/install.sh')
          - 'pyinfra' scripts must end with '.py' and 'pyinfra' must be installed.
            See https://pyinfra.readthedocs.io/en/latest/ for more info.
          - Provisioning is run when the instance is started. This option
            is if you want to re-run the provisioning.
          - An example of provisioning could be installing puppet (or your config tool
            of choice).

        Options:
            -h, --help                       Print this help
            -s, --show-only                  Show the provisioning info (do not run)
        """
        show = arguments['--show-only']
        instance_name = arguments['<instance>']

        if instance_name:
            # single instance
            instances = [instance_name]
        else:
            # multiple instances
            instances = self.instances()

        for instance in instances:
            inst = MechInstance(instance)

            if inst.created:
                if inst.provider == 'vmware':
                    utils.provision(inst, show)
                else:
                    print(colored.red("Not yet implemented on this platform."))
            else:
                print("VM not created.")

    def reload(self, arguments):
        """
        Restarts Mech machine, loads new Mechfile configuration.

        Usage: mech reload [options] [<instance>]

        Options:
            -h, --help                       Print this help
        """
        instance_name = arguments['<instance>']

        if instance_name:
            # single instance
            instances = [instance_name]
        else:
            # multiple instances
            instances = self.instances()

        for instance in instances:
            inst = MechInstance(instance)

            if inst.created:

                if inst.provider == 'vmware':

                    vmrun = VMrun(inst.vmx)

                    print(colored.blue("Reloading machine..."))
                    started = vmrun.reset()
                    if started is None:
                        print(colored.red("VM not restarted"))
                    else:
                        print(colored.blue("Getting IP address..."))
                        ip_address = inst.get_ip(wait=True)
                        if ip_address:
                            if started:
                                print(colored.green("VM ({}) started "
                                                    "on {}".format(instance, ip_address)))
                            else:
                                print(colored.yellow("VM ({}) already was started on "
                                                     "{}".format(instance, ip_address)))
                        else:
                            if started:
                                print(colored.green("VM ({}) started on an unknown IP "
                                                    "address".format(instance)))
                            else:
                                print(colored.yellow("VM ({}) already was started "
                                                     "on an unknown IP address".format(instance)))
                else:
                    print(colored.red("Not yet implemented on this platform."))
            else:
                print("VM not created.")

    def port(self, arguments):
        """
        Displays information about guest port mappings.

        Usage: mech port [options] [<instance>]

        Options:
                --guest PORT                 Output the host port that maps to the given guest port
            -h, --help                       Print this help
        """
        instance_name = arguments['<instance>']

        if platform.system() == 'Linux':
            sys.exit('This command is not supported on this OS.')

        if instance_name:
            # single instance
            instances = [instance_name]
        else:
            # multiple instances
            instances = self.instances()

        # FUTURE: implement port forwarding?
        for instance in instances:
            inst = MechInstance(instance)

            if inst.provider == 'vmware':
                print('Instance ({}):'. format(instance))
                nat_found = False
                vmrun = VMrun(inst.vmx)
                for line in vmrun.list_host_networks().split('\n'):
                    network = line.split()
                    if len(network) > 2 and network[2] == 'nat':
                        print(vmrun.list_port_forwardings(network[1]))
                        nat_found = True
                if not nat_found:
                    print(colored.red("Cannot find a nat network"), file=sys.stderr)
            else:
                print(colored.red("Not yet implemented on this platform."))

    def list(self, arguments):
        """
        Lists all available boxes from Mechfile.

        Usage: mech list [options]

        Options:
            -d, --detail                     Print detailed info
            -h, --help                       Print this help
        """

        detail = arguments['--detail']

        self.activate_mechfile()

        if detail:
            print('Instance Details')
            print()
        else:
            print("{}\t{}\t{}\t{}\t{}".format(
                'NAME'.rjust(20),
                'ADDRESS'.rjust(15),
                'BOX'.rjust(35),
                'VERSION'.rjust(12),
                'PROVIDER'.rjust(12),
            ))

        for name in self.mechfile:
            inst = MechInstance(name, self.mechfile)
            if inst.created:
                ip_address = inst.get_ip()
                if ip_address is None:
                    ip_address = colored.yellow("poweroff")
                elif not ip_address:
                    ip_address = colored.green("running")
                else:
                    ip_address = colored.green(ip_address)
            else:
                ip_address = "notcreated"

            if detail:
                print(inst)
                print()
            else:
                # deal with box_version being none
                box_version = inst.box_version
                if inst.box_version is None:
                    box_version = ''
                print("{}\t{}\t{}\t{}\t{}".format(
                    colored.green(name.rjust(20)),
                    ip_address.rjust(15),
                    inst.box.rjust(35),
                    box_version.rjust(12),
                    inst.provider.rjust(12)
                ))

    # allow 'mech ls' as alias to 'mech list'
    ls = list



# Copyright (c) 2020 Mike Kinney

"""Unit tests for 'mech box'."""
import re

from unittest.mock import patch

import mech.command
import mech.mech
import mech.vmrun


@patch('os.getcwd')
def test_mech_box_list_no_mechdir(mock_os_getcwd, capfd):
    """Test 'mech box list' with no '.mech' directory."""
    mock_os_getcwd.return_value = '/tmp'
    global_arguments = {'--debug': False}
    a_mech = mech.mech.MechBox(arguments=global_arguments)
    with patch('os.walk') as mock_walk:
        # root, dirs, files
        mock_walk.return_value = [('./tmp', [], []), ]
        a_mech.list({})
        mock_walk.assert_called()
        out, _ = capfd.readouterr()
        # ensure a header prints out
        assert re.search(r'BOX', out, re.MULTILINE)


@patch('os.getcwd')
def test_mech_box_list_empty_boxes_dir(mock_os_getcwd, capfd):
    """Test 'mech box list' with no directories in '.mech/boxes' directory."""
    mock_os_getcwd.return_value = '/tmp'
    global_arguments = {'--debug': False}
    a_mech = mech.mech.MechBox(arguments=global_arguments)
    with patch('os.walk') as mock_walk:
        # root, dirs, files
        mock_walk.return_value = [('/tmp', ['boxes', ], []), ]
        a_mech.list({})
        mock_walk.assert_called()
        out, _ = capfd.readouterr()
        # ensure a header prints out
        assert re.search(r'BOX', out, re.MULTILINE)


@patch('os.getcwd')
def test_mech_box_list_one_box(mock_os_getcwd, capfd):
    """Test 'mech box list' with one box present."""
    mock_os_getcwd.return_value = '/tmp'
    global_arguments = {'--debug': False}
    a_mech = mech.mech.MechBox(arguments=global_arguments)
    with patch('os.walk') as mock_walk:
        # simulate: bento/ubuntu-18.04/201912.04.0/vmware_desktop.box
        mock_walk.return_value = [
            ('/tmp', ['boxes'], []),
            ('/tmp/boxes', ['bento'], []),
            ('/tmp/boxes/vmware/bento', ['ubuntu-18.04'], []),
            ('/tmp/boxes/vmware/bento/ubuntu-18.04', ['201912.04.0'], []),
            ('/tmp/boxes/vmware/bento/ubuntu-18.04/201912.04.0', [], ['vmware_desktop.box']),
        ]
        a_mech.list({})
        mock_walk.assert_called()
        out, _ = capfd.readouterr()
        assert re.search(r'ubuntu-18.04', out, re.MULTILINE)


@patch('requests.get')
@patch('os.path.exists')
@patch('os.getcwd')
def test_mech_box_add_new(mock_os_getcwd, mock_os_path_exists,
                          mock_requests_get, capfd, catalog_as_json,
                          mech_box_arguments):
    """Test 'mech box add' from Hashicorp'."""
    mock_os_path_exists.return_value = False
    mock_os_getcwd.return_value = '/tmp'
    global_arguments = {'--debug': False}
    mock_requests_get.return_value.status_code = 200
    mock_requests_get.return_value.json.return_value = catalog_as_json

    a_mech = mech.mech.MechBox(arguments=global_arguments)
    arguments = mech_box_arguments
    arguments['<location>'] = 'bento/ubuntu-19.10'
    arguments['<provider>'] = 'vmware'
    a_mech.add(arguments)
    out, _ = capfd.readouterr()
    assert re.search(r'Checking integrity', out, re.MULTILINE)


@patch('requests.get')
@patch('os.path.exists')
@patch('os.getcwd')
def test_mech_box_add_existing(mock_os_getcwd, mock_os_path_exists,
                               mock_requests_get, capfd, catalog_as_json,
                               mech_box_arguments):
    """Test 'mech box add' from Hashicorp'."""
    mock_os_getcwd.return_value = '/tmp'
    mock_os_path_exists.return_value = True
    global_arguments = {'--debug': False}
    mock_requests_get.return_value.status_code = 200
    mock_requests_get.return_value.json.return_value = catalog_as_json

    a_mech = mech.mech.MechBox(arguments=global_arguments)
    arguments = mech_box_arguments
    arguments['<location>'] = 'bento/ubuntu-19.10'
    a_mech.add(arguments)
    out, _ = capfd.readouterr()
    assert re.search(r'Loading metadata', out, re.MULTILINE)


@patch('shutil.rmtree')
@patch('os.path.exists')
def test_mech_box_remove_exists(mock_os_path_exists, mock_rmtree, capfd):
    """Test 'mech box remove'."""
    mock_os_path_exists.return_value = True
    mock_rmtree.return_value = True
    global_arguments = {'--debug': False}
    a_mech = mech.mech.MechBox(arguments=global_arguments)
    arguments = {
        '<name>': 'bento/ubuntu-18.04',
        '<version>': 'somever',
        '<provider>': 'vmware',
    }
    a_mech.remove(arguments)
    out, _ = capfd.readouterr()
    mock_os_path_exists.assert_called()
    mock_rmtree.assert_called()
    assert re.search(r'Removed ', out, re.MULTILINE)


@patch('os.path.exists')
def test_mech_box_remove_does_not_exists(mock_os_path_exists, capfd):
    """Test 'mech box remove'."""
    mock_os_path_exists.return_value = False
    global_arguments = {'--debug': False}
    a_mech = mech.mech.MechBox(arguments=global_arguments)
    arguments = {
        '<name>': 'bento/ubuntu-18.04',
        '<version>': 'somever',
        '<provider>': 'vmware',
    }
    a_mech.remove(arguments)
    out, _ = capfd.readouterr()
    mock_os_path_exists.assert_called()
    assert re.search(r'No boxes were removed', out, re.MULTILINE)



# Copyright (c) 2020 Mike Kinney

"""Mech integration tests: multiple instances"""
import re
import subprocess


import pytest


@pytest.mark.int
def test_int_multiple_instances(helpers):
    """Test with multiple instances."""

    test_dir = "tests/int/multiple_instances"
    helpers.cleanup_dir_and_vms_from_dir(test_dir)

    mechfile_contents = """{
  "first": {
    "box": "bento/ubuntu-18.04",
    "box_version": "201912.04.0",
    "name": "first",
    "url": "https://vagrantcloud.com/bento/boxes/ubuntu-18.04/\
versions/201912.04.0/providers/vmware_desktop.box"
  },
  "second": {
    "box": "bento/ubuntu-18.04",
    "box_version": "201912.04.0",
    "name": "second",
    "url": "https://vagrantcloud.com/bento/boxes/ubuntu-18.04/\
versions/201912.04.0/providers/vmware_desktop.box"
  }
}
    """
    mechfile_path = test_dir + '/' + 'Mechfile'
    mechfile_file = open(mechfile_path, "w")
    mechfile_file.write(mechfile_contents)
    mechfile_file.close()

    # list two
    command = "mech ls"
    expected_lines = ["first", "second"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # should start
    command = "mech up"
    expected_lines = ["could not be found", "vmware_desktop",
                      "integrity", "Extracting", "Added network",
                      "Bringing machine", "Getting IP", "Sharing folders",
                      "started", "Provisioning"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    command = "mech add third bento/ubuntu-18.04"
    expected_lines = ["Adding", "Loading", "Added"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # should stop one
    command = "mech stop first"
    expected_lines = ["Stopped"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # should only have one instance running (second)
    command = "mech ls"
    expected_lines = [r"first.*poweroff",
                      r"second.*[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}",
                      r"third.*notcreated"]
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    for line in expected_lines:
        print(line)
        assert re.search(line, stdout, re.MULTILINE)

    # should be able to destroy
    command = "mech destroy -f"
    expected = "Deleting"
    results = subprocess.run(command, cwd=test_dir, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    assert stderr == ''
    assert results.returncode == 0
    assert re.search(expected, stdout)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016-2017 Kevin Chung
# Copyright (c) 2018 German Mendez Bravo (Kronuz)
# Copyright (c) 2020 Mike Kinney
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
"""Compatibility code for different versions of python."""
import sys
import operator
import functools

#: Python 3 detection
PY3 = sys.version_info[0] == 3

# Get a reference to the builtins module
try:
    _builtins_ = __import__('builtins')
except ImportError:
    _builtins_ = __import__('__builtin__')

#: Singlular builtin accessor
_builtin = functools.partial(getattr, _builtins_)

#: No operation lambda


def _noop(obj):
    return obj


if PY3:
    meth_func = '__func__'
    raw_input = _builtin('input')

    binary_type = _builtin('bytes')

    # str <-> bytes conversions
    s = operator.methodcaller('decode', 'latin-1')
    b = operator.methodcaller('encode', 'latin-1')

else:
    meth_func = 'im_func'
    raw_input = _builtin('raw_input')

    binary_type = _builtin('str')

    # str <-> bytes conversions (not necessary in python 2)
    s = _noop
    b = _noop


def o(numstr):
    """Octal number compatibility shim."""
    return int(numstr, 8)


def get_meth_func(klass):
    """Get a bound method's function."""
    return operator.attrgetter(
        meth_func)(klass) if hasattr(klass, meth_func) else None


def b2s(bytestr):
    '''"safe" form of ``b``. Checks for binary type before operating.'''
    return s(bytestr) if isinstance(bytestr, binary_type) else bytestr
